import threading
import time, datetime
import json
import logging
import sys, traceback
from flashlexiot.sdk import FlashlexSDK

from AWSIoTPythonSDK.MQTTLib import AWSIoTMQTTClient

LOG_FORMAT = ('%(levelname) -10s %(asctime)s %(funcName) '
              '-35s %(lineno) -5d: %(message)s')

LOGGER = logging.getLogger(__name__)

# init LOGGER
logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)


class ThreadTypeFactory:
    def get_thread_for_config(self, config, customCallback):
        if config["flashlex"]["app"]["thread"] == 'basicPubsub':
            return BasicPubsubThread(config, customCallback)
        elif config["flashlex"]["app"]["thread"] == 'subscribe':
            return TopicSubscribeThread(config, customCallback)            
        elif config["flashlex"]["app"]["thread"] == 'test':
            return TestPublishThread(config, customCallback)            
        else:
            raise ValueError(config["flashlex"]["app"]["callback"])

class ExpireMessagesThread(threading.Thread):
    def __init__(self, config):
        super(ExpireMessagesThread, self).__init__()
        self._expires = config["flashlex"]["app"]["db"]["expireSeconds"]
        self._dbpath = "{0}/{1}".format(
            config["flashlex"]["app"]["db"]["dataPath"], 
            config["flashlex"]["app"]["db"]["subscriptionData"])
        self._sdk = FlashlexSDK(config)


    def run(self):
        LOGGER.info("running message cleanup")   
        if(self._expires>0):
            threshold = datetime.datetime.fromtimestamp(time.mktime(time.gmtime()))-datetime.timedelta(seconds=int(self._expires))
            ts = threshold.strftime("%s")  

            messages = self._sdk.getSubscribedMessages()
            for message in messages:
                if 'timestamp' in message and message['timestamp'] < float(ts):
                    self._sdk.removeMessageFromStore(message)
    

class TopicSubscribeThread(threading.Thread):
    def __init__(self, config, customCallback):
        super(TopicSubscribeThread, self).__init__()
        self._config = config
        self.thingName = config["flashlex"]["thing"]["name"]
        self._customCallback = customCallback
        self._iotClient = setupClientFromConfig(config)
        self.topic = config["flashlex"]["thing"]["mqtt"]["ingress"]["topic"]
        self._type = "subscribe"

    def getType(self):
        return self._type    

    def run(self):
        """ save messages subscribed to to 
        local database
        """
        #topic="ingress/{0}".format(self.thingName)

        # Connect and subscribe to AWS IoT
        self._iotClient.connect()
        self._iotClient.subscribe(self.topic, 1, self._customCallback)
        time.sleep(4)

        loop = True
        while loop:
            # LOGGER.info("subscribe-{0} listening to topic: {1}".format(
            #      self.thingName, self.topic))
            time.sleep(10)

class BasicPubsubThread(threading.Thread):
    def __init__(self, config, customCallback):
        super(BasicPubsubThread, self).__init__()    
        self._config = config
        self.message = config["flashlex"]["app"]["message"]
        self.thingName = config["flashlex"]["thing"]["name"]
        self._customCallback = customCallback
        self._iotClient = setupClientFromConfig(config)
        self.topic = config["flashlex"]["thing"]["mqtt"]["pubsub"]["topic"]
        self._type = "basicPubsub"
        self._loopUntil = -1

        if("loopCount" in config["flashlex"]["app"]):
            self._loopUntil = config["flashlex"]["app"]["loopCount"]

    def getType(self):
        return self._type

    def run(self):
        """ save messages subscribed to to 
        local database
        """
        # print("running")
        # topic="pubsub/{0}".format(self.thingName)

        # Connect and subscribe to AWS IoT
        self._iotClient.connect()
        self._iotClient.subscribe(self.topic, 1, self._customCallback)
        time.sleep(2)    

        # Publish to the same topic in a loop forever
        loopCount = 0
        loop = True

        while loop:
            try:
                messageModel = {}
                messageModel['message'] = self.message
                messageModel['sequence'] = loopCount
                messageJson = json.dumps(messageModel)
                self._iotClient.publish(self.topic, messageJson, 1)
                LOGGER.info('Published topic %s: %s\n' % (self.topic, messageJson))
                loopCount += 1
                time.sleep(5)
                if(self._loopUntil>0 and loopCount>=self._loopUntil):
                    loop = False

            except:
                LOGGER.error("an error occured.")
                print("-"*60)
                traceback.print_exc(file=sys.stdout)
                print("-"*60)
                loop = False


class TestPublishThread(threading.Thread):
    def __init__(self, config, customCallback):
        super(TestPublishThread, self).__init__()    
        self._config = config
        self.message = config["flashlex"]["app"]["message"]
        self.thingName = config["flashlex"]["thing"]["name"]
        self.thingId = config["flashlex"]["thing"]["id"]
        self._customCallback = customCallback
        self._iotClient = setupClientFromConfig(config)
        self.topic = "flashlex/test"
        self._type = "test"
        self._loopUntil = 10

        if("loopCount" in config["flashlex"]["app"]):
            self._loopUntil = config["flashlex"]["app"]["loopCount"]
        
    def run(self):
        """ save messages subscribed to to 
        local database
        """
        # Connect and subscribe to AWS IoT
        self._iotClient.connect()
        time.sleep(2)    


        # Publish to the same topic in a loop forever
        loopCount = 0
        for loopCount in range(self._loopUntil):    
            try:
                messageModel = {}
                messageModel['thingId'] = self.thingId
                messageModel['thingName'] = self.thingName
                messageModel['message'] = self.message
                messageModel['sequence'] = loopCount
                messageJson = json.dumps(messageModel)
                self._iotClient.publish(self.topic, messageJson, 1)
                LOGGER.info('Published topic %s: %s\n' % (self.topic, messageJson))
                loopCount += 1
                time.sleep(5)

            except:
                LOGGER.error("an error occured.")
                print("-"*60)
                traceback.print_exc(file=sys.stdout)
                print("-"*60)
                loop = False


def setupClientFromConfig(config):
    return setupClient(
        config["flashlex"]["thing"]["name"], 
        config["flashlex"]["thing"]["endpoint"], 
        config["flashlex"]["thing"]["mqtt"]["port"], 
        "{0}/{1}".format(config["flashlex"]["thing"]["keys"]["path"], config["flashlex"]["thing"]["keys"]["rootCA"]), #rootca
        "{0}/{1}".format(config["flashlex"]["thing"]["keys"]["path"], config["flashlex"]["thing"]["keys"]["privateKey"]), #private key
        "{0}/{1}".format(config["flashlex"]["thing"]["keys"]["path"], config["flashlex"]["thing"]["keys"]["cert"]), #cert 
        config["flashlex"]["thing"]["mqtt"]["useWebsocket"],
        config["flashlex"]["thing"]["mqtt"]["autoReconnectBackoffTime"]["baseReconnectQuietTimeSecond"],
        config["flashlex"]["thing"]["mqtt"]["autoReconnectBackoffTime"]["maxReconnectQuietTimeSecond"],
        config["flashlex"]["thing"]["mqtt"]["autoReconnectBackoffTime"]["stableConnectionTimeSecond"], 
        config["flashlex"]["thing"]["mqtt"]["offlinePublishQueueing"],
        config["flashlex"]["thing"]["mqtt"]["drainingFrequency"],
        config["flashlex"]["thing"]["mqtt"]["connectDisconnectTimeout"],
        config["flashlex"]["thing"]["mqtt"]["mqttOperationTimeout"])

def setupClient(
    clientId, 
    host, port, 
    rootCAPath, 
    privateKeyPath, 
    certificatePath, 
    useWebsocket, 
    baseReconnectQuietTimeSecond,
    maxReconnectQuietTimeSecond,
    stableConnectionTimeSecond,
    offlinePublishQueueing,
    drainingFrequency,
    connectDisconnectTimeout,
    mqttOperationTimeout
    ):

    # Init AWSIoTMQTTClient
    myAWSIoTMQTTClient = None
    if useWebsocket:
        myAWSIoTMQTTClient = AWSIoTMQTTClient(clientId, useWebsocket=True)
        myAWSIoTMQTTClient.configureEndpoint(host, port)
        myAWSIoTMQTTClient.configureCredentials(rootCAPath)
    else:
        myAWSIoTMQTTClient = AWSIoTMQTTClient(clientId)
        myAWSIoTMQTTClient.configureEndpoint(host, port)
        myAWSIoTMQTTClient.configureCredentials(rootCAPath, privateKeyPath, certificatePath)

    # AWSIoTMQTTClient connection configuration
    myAWSIoTMQTTClient.configureAutoReconnectBackoffTime( 
        baseReconnectQuietTimeSecond,
        maxReconnectQuietTimeSecond,
        stableConnectionTimeSecond)
    myAWSIoTMQTTClient.configureOfflinePublishQueueing(offlinePublishQueueing)  # Infinite offline Publish queueing
    myAWSIoTMQTTClient.configureDrainingFrequency(drainingFrequency)  # Draining: 2 Hz
    myAWSIoTMQTTClient.configureConnectDisconnectTimeout(connectDisconnectTimeout)  # 10 sec
    myAWSIoTMQTTClient.configureMQTTOperationTimeout(mqttOperationTimeout)  # 5 sec
 
    return myAWSIoTMQTTClient

threadTypeFactory = ThreadTypeFactory()    



import uuid
import time
import datetime
import json

from flashlexiot.sdk import FlashlexSDK

class CallbackFactory:
    def get_callback_for_config(self, config):
        if config["flashlex"]["app"]["callback"] == 'basicPubsub':
            return BasicPubsubCallbackHandler(config)
        elif config["flashlex"]["app"]["callback"] == 'persistent':
            return PersistentCallbackHandler(config)  
        elif config["flashlex"]["app"]["callback"] == 'test':
            return TestCallbackHandler(config)                    
        else:
            raise ValueError(config["flashlex"]["app"]["callback"])

class BasicPubsubCallbackHandler(object):

    def __init__(self, config):
        self._client = config["flashlex"]["thing"]["name"]
        self._type = "basicPubsub"

    def handleMessage(self, client, userdata, message):
        print("Received a new message on client:{0} type:{1}: ".format(self._client, self._type))
        print(message.payload.decode("utf-8"))
        print("from topic: ")
        print(message.topic)
        print("--------------\n\n")

class TestCallbackHandler(object):

    def __init__(self, config):
        self._client = config["flashlex"]["thing"]["name"]
        self._type = "test"

    def handleMessage(self, client, userdata, message):
        print("Received a new message on client:{0} type:{1}: ".format(self._client, self._type))
        print(message.payload.decode("utf-8"))
        print("from topic: ")
        print(message.topic)
        print("--------------\n\n")


class PersistentCallbackHandler(object):

    def __init__(self, config):
        self._client = config["flashlex"]["thing"]["name"] 
        self._dbpath = "{0}/{1}".format(
            config["flashlex"]["app"]["db"]["dataPath"], 
            config["flashlex"]["app"]["db"]["subscriptionData"])

        self._type = "persistent"

        self._sdk = FlashlexSDK(config)

    def handleMessage(self, client, userdata, message):

        ts= time.time()
        messageDoc = {}
        messageDoc["pk"] = str(uuid.uuid4()).replace("-","")[:12]
        messageDoc["timestamp"] = ts
        messageDoc["datetime"] = "{0}".format(datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S'))        
        messageDoc["message"] = {
                "topic" : message.topic,
                "payload": json.loads(message.payload.decode("utf-8")),
                "pos": message.qos,
                "retain": message.retain,
                "mid": message.mid
        }
        
        self._sdk.setMessageToStore(messageDoc)

        print("Received a new message on client:{0} type:{1}: ".format(self._client, self._type))
        print(message.payload.decode("utf-8"))
        print("from topic: ")
        print(message.topic)
        print("--------------\n\n")

callbackFactory = CallbackFactory()





"""
circleci.version
~~~~~~~~~~~~~~~~
"""
VERSION = "0.9.5"
"""Current version of circleci.py."""

def get_short_version():
    """
    Format "short" version in the form of X.Y
    :returns: short version
    """
    ver = VERSION.split(".")
    short_version = "{0}.{1}".format(ver[0], ver[1])
    return short_version



# -*- coding: utf-8 -*-
"""
circleci.error
~~~~~~~~~~~~~~

    This module provides some classes that subclass Exception to provide
    some useful exceptions for the API wrapper.
"""


class CircleCIException(Exception):
    """Base class for CircleCI exceptions

    :param argument: The argument that was passed into the function.
    """
    def __init__(self, argument):
        super().__init__()
        self.argument = argument
        self.message = None

    def __str__(self):
        return '{0} is invalid. {1}'.format(self.argument, self.message)


class BadVerbError(CircleCIException):
    """Exception raises for bad HTTP verb

    :param argument: The argument that was passed into the function.
    """
    message = "verb must be one of 'GET', 'POST', or 'DELETE'"

    def __init__(self, argument):
        super().__init__(argument)
        self.message = BadVerbError.message


class BadKeyError(CircleCIException):
    """Exception raises for bad Key Type

    :param argument: The argument that was passed into the function.
    """
    message = "key must be one of 'deploy-key' or 'github-user-key'"

    def __init__(self, argument):
        super().__init__(argument)
        self.message = BadKeyError.message


class InvalidFilterError(CircleCIException):
    """Exception raises for an invalid filter

    :param argument: The argument that was passed into the function.
    :param filter_type: Filter for status or artifacts.
    """
    filter_message = "status_filter must be one of 'completed'" \
        "'successful', 'failed', or 'running'"
    artifacts_message = "must be one of 'completed', 'successful', or 'failed'"

    def __init__(self, argument, filter_type):
        super().__init__(argument)
        if filter_type == 'status':
            self.message = InvalidFilterError.filter_message
        if filter_type == 'artifacts':
            self.message = InvalidFilterError.artifacts_message


import unittest
import pathlib
import os, sys, traceback
import yaml
import pickledb

from os.path import dirname, abspath
from shutil import copyfile
from flashlexiot.backend.thread import BasicPubsubThread, ExpireMessagesThread
from flashlexiot.sdk import FlashlexSDK

def loadConfig(configFile):
    cfg = None
    with open(configFile, 'r') as ymlfile:
        cfg = yaml.load(ymlfile, Loader=yaml.FullLoader)
    return cfg

class TestFlashlexSDK(unittest.TestCase):

    def setUp(self):
        fn = pathlib.Path(__file__).parent / 'test-config.yml'

         #get defaults for data and keys
        dir_path = os.path.dirname(os.path.realpath(__file__))
        config = loadConfig("{0}/test-config.yml".format(dir_path))

        self.sdk = FlashlexSDK(config)

        config = self.sdk.getConfig()
        config["flashlex"]["app"]["db"]["dataPath"] = pathlib.Path(__file__).parent 
        config["flashlex"]["app"]["db"]["subscriptionData"] = 'data/subscription2.db'
        self.sdk.setConfig(config)

        message1 = {
            "pk": "b300d03c-7bd4-4110-9489-5ef59abb1981",
            "timestamp": 1553978975.5232847,
            "datetime": "2019-03-30 13:49:35",
            "message": {
                "topic": "pubsub/foobar30",
                "payload": {
                    "message": "Sending a basic message...",
                    "sequence": 151
                },
                "pos": 1,
                "retain": 0,
                "mid": 1
            }
        }
        
        message2 = {
            "pk": "a300d03c-7bd4-4110-9489-5ef59abb1981",
            "timestamp": 1553978976.5232847,
            "datetime": "2019-03-30 13:49:35",
            "message": {
                "topic": "pubsub/foobar30",
                "payload": {
                    "message": "Sending a basic message...",
                    "sequence": 151
                },
                "pos": 1,
                "retain": 0,
                "mid": 1
            }
        }

        subscriptionDataPath = "{0}/{1}".format(
            config["flashlex"]["app"]["db"]["dataPath"], 
            config["flashlex"]["app"]["db"]["subscriptionData"])

        subscriptionDb = pickledb.load(subscriptionDataPath, False)
        subscriptionDb.set(message1['pk'], message1)
        subscriptionDb.set(message2['pk'], message2)
        subscriptionDb.dump()

    def test_load_config(self):
        self.assertEqual('testThing1', self.sdk.getConfig()["flashlex"]["thing"]["name"])

    def test_get_messages(self):
        messages = self.sdk.getSubscribedMessages()
        self.assertEqual(2,len(messages))

    def test_remove_message(self):
        config = self.sdk.getConfig()
        messages = self.sdk.getSubscribedMessages()
        self.sdk.removeMessageFromStore(messages[0])
        messages = self.sdk.getSubscribedMessages()
        self.assertEqual(1,len(messages))

    def tearDown(self):
        print('tear down')

        config = self.sdk.getConfig()
        subscriptionDataPath = "{0}/{1}".format(
            config["flashlex"]["app"]["db"]["dataPath"], 
            config["flashlex"]["app"]["db"]["subscriptionData"])

        os.remove(subscriptionDataPath)



if __name__ == '__main__':
    unittest.main()



import os
import sys
import setuptools

from setuptools import setup
from setuptools.command.install import install
# from circleci.version import VERSION

with open("README.md", "r") as fh:
    long_description = fh.read()


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')

        if tag != '0.9.15':
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, '0.9.15'
            )
            sys.exit(info)

setuptools.setup(
    name="flashlexiot",
    version="0.9.15",
    author="Clay Graham",
    author_email="claytantor@flashlex.com",
    description="Flashlex IOT for python makes it easy to make any python computer an IOT device.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/claytantor/flashlex-iot-python",
    packages=setuptools.find_packages(),
    classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", 
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6'
    ],
    install_requires=[
        'requests',
    ],
    python_requires='>=3.4',
    cmdclass={
        'verify': VerifyVersionCommand,
    }

)



# -*- coding: utf-8 -*-
"""
circleci.api
~~~~~~~~~~~~

    This module provides a class which abstracts the CircleCI REST API.
"""
import os

import requests
from requests.auth import HTTPBasicAuth

from circleci.error import BadKeyError, BadVerbError, InvalidFilterError


class Api():
    """A python interface into the CircleCI API"""

    def __init__(self, token, url='https://circleci.com/api/v1.1'):
        """Instantiate a new circleci.Api object.

        :param url: The URL to the CircleCI instance. Defaults to \
            https://circleci.com/api/v1.1. If you are running CircleCI server, \
            the API is available at the same endpoint of your own \
            installation url. i.e (https://circleci.yourcompany.com/api/v1.1).
        :param token: Your CircleCI API token.
        """
        self.token = token
        self.url = url

    def get_user_info(self):
        """Provides information about the signed in user.

        Endpoint:
            GET: ``/me``
        """
        resp = self._request('GET', 'me')
        return resp

    def get_projects(self):
        """List of all the projects you're following on CircleCI.

        Endpoint:
            GET: ``/projects``
        """
        resp = self._request('GET', 'projects')
        return resp

    def follow_project(self, username, project, vcs_type='github'):
        """Follow a new project on CircleCI.

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        Endpoint:
            POST: ``/project/:vcs-type/:username/:project/follow``
        """
        endpoint = 'project/{0}/{1}/{2}/follow'.format(
            vcs_type,
            username,
            project
        )
        resp = self._request('POST', endpoint)
        return resp

    def get_project_build_summary(
            self,
            username,
            project,
            limit=30,
            offset=0,
            status_filter=None,
            branch=None,
            vcs_type='github'):
        """Build summary for each of the last 30 builds for a single git repo.

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param limit: The number of builds to return. Maximum 100, defaults \
            to 30.
        :param offset: The API returns builds starting from this offset, \
            defaults to 0.
        :param status_filter: Restricts which builds are returned. \
            Set to "completed", "successful", "running" or "failed". \
            Defaults to no filter.
        :param branch: Narrow returned builds to a single branch.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        :type limit: int
        :type offset: int

        :raises InvalidFilterError: when filter is not a valid filter.

        Endpoint:
            GET: ``/project/:vcs-type/:username/:project``
        """
        valid_filters = [None, 'completed', 'successful', 'failed', 'running']

        if status_filter not in valid_filters:
            raise InvalidFilterError(status_filter, 'status')

        if branch:
            endpoint = 'project/{0}/{1}/{2}/tree/{3}?limit={4}&offset={5}&filter={6}'.format(
                vcs_type,
                username,
                project,
                branch,
                limit,
                offset,
                status_filter
            )
        else:
            endpoint = 'project/{0}/{1}/{2}?limit={3}&offset={4}&filter={5}'.format(
                vcs_type,
                username,
                project,
                limit,
                offset,
                status_filter
            )

        resp = self._request('GET', endpoint)
        return resp

    def get_recent_builds(self, limit=30, offset=0):
        """
        Build summary for each of the last 30 recent builds, ordered by build_num.

        :param limit: The number of builds to return. Maximum 100, defaults \
            to 30.
        :param offset: The API returns builds starting from this offset, \
            defaults to 0.

        :type limit: int
        :type offset: int

        Endpoint:
            GET: ``/recent-builds``
        """
        endpoint = 'recent-builds?limit={0}&offset={1}'.format(limit, offset)

        resp = self._request('GET', endpoint)
        return resp

    def get_build_info(self, username, project, build_num, vcs_type='github'):
        """Full details for a single build.

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param build_num: Build number.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        Endpoint:
            GET: ``/project/:vcs-type/:username/:project/:build_num``
        """
        endpoint = 'project/{0}/{1}/{2}/{3}'.format(
            vcs_type,
            username,
            project,
            build_num
        )
        resp = self._request('GET', endpoint)
        return resp

    def get_artifacts(self, username, project, build_num, vcs_type='github'):
        """List the artifacts produced by a given build.

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param build_num: Build number.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        Endpoint:
            GET: ``/project/:vcs-type/:username/:project/:build_num/artifacts``
        """
        endpoint = 'project/{0}/{1}/{2}/{3}/artifacts'.format(
            vcs_type,
            username,
            project,
            build_num
        )
        resp = self._request('GET', endpoint)
        return resp

    def get_latest_artifact(
            self,
            username,
            project,
            branch=None,
            status_filter='completed',
            vcs_type='github'):
        """List the artifacts produced by the latest build on a given branch.

        .. note::
            This endpoint is a little bit flakey. If the "latest" \
                build does not have any artifacts, rathern than returning \
                an empty set, the API will 404.

        :param username: org or user name
        :param project: case sensitive repo name
        :param branch: The branch you would like to look in for the latest build.
                Returns artifacts for latest build in entire project if omitted.
        :param filter: Restricts which builds are returned.
                defaults to 'completed'
                valid filters: "completed", "successful", "failed"
        :param vcs_type: defaults to github
                on circleci.com you can also pass in bitbucket

        :raises InvalidFilterError: when filter is not a valid filter.

        Endpoint:
            GET: ``/project/:vcs-type/:username/:project/latest/artifacts``
        """
        valid_filters = ['completed', 'successful', 'failed']

        if status_filter not in valid_filters:
            raise InvalidFilterError(status_filter, 'artifacts')

        # passing None makes the API 404
        if branch:
            endpoint = 'project/{0}/{1}/{2}/latest/artifacts?branch={3}&filter={4}'.format(
                vcs_type,
                username,
                project,
                branch,
                status_filter
            )
        else:
            endpoint = 'project/{0}/{1}/{2}/latest/artifacts?filter={3}'.format(
                vcs_type,
                username,
                project,
                status_filter
            )

        resp = self._request('GET', endpoint)
        return resp

    def download_artifact(self, url, destdir=None, filename=None):
        """Download an artifact from a url

        :param url: The URL to the artifact.
        :param destdir: The optional destination directory. \
            Defaults to None (curent working directory).
        :param filename: Optional file name. Defaults to the name of the artifact file.
        """
        resp = self._download(url, destdir, filename)
        return resp

    def retry_build(self, username, project, build_num, ssh=False, vcs_type='github'):
        """Retries the build.

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param build_num: Build number.
        :param ssh: Retry a build with SSH enabled. Defaults to False.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        :type ssh: bool

        Endpoint:
            POST: ``/project/:vcs-type/:username/:project/:build_num/retry``
        """
        if ssh:
            endpoint = 'project/{0}/{1}/{2}/{3}/ssh'.format(
                vcs_type,
                username,
                project,
                build_num
            )
        else:
            endpoint = 'project/{0}/{1}/{2}/{3}/retry'.format(
                vcs_type,
                username,
                project,
                build_num
            )
        resp = self._request('POST', endpoint)
        return resp

    def cancel_build(self, username, project, build_num, vcs_type='github'):
        """Cancels the build.

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param build_num: Build number.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        Endpoint:
            POST: ``/project/:vcs-type/:username/:project/:build_num/cancel``
        """
        endpoint = 'project/{0}/{1}/{2}/{3}/cancel'.format(
            vcs_type,
            username,
            project,
            build_num
        )
        resp = self._request('POST', endpoint)
        return resp

    def add_ssh_user(self, username, project, build_num, vcs_type='github'):
        """Adds a user to the build's SSH permissions.

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param build_num: Build number.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        Endpoint:
            POST: ``/project/:vcs-type/:username/:project/:build_num/ssh-users``
        """
        endpoint = 'project/{0}/{1}/{2}/{3}/ssh-users'.format(
            vcs_type,
            username,
            project,
            build_num
        )
        resp = self._request('POST', endpoint)
        return resp


    def trigger_build(
            self,
            username,
            project,
            branch='master',
            revision=None,
            tag=None,
            parallel=None,
            params=None,
            vcs_type='github'):
        """
        Triggers a new build.

        .. note::
            * ``tag`` and ``revision`` are mutually exclusive.
            * ``parallel`` is ignored for builds running on CircleCI 2.0

        :param username: Organization or user name.
        :param project: Case sensitive repo name.
        :param branch: The branch to build. Defaults to master.
        :param revision: The specific git revision to build. \
            Default is null and the head of the branch is used. \
            Can not be used with the tag parameter.
        :param tag: The git tag to build. \
            Default is null. \
            Cannot be used with the tag parameter.
        :param parallel: The number of containers to use to run the build. \
                Default is null and the project default is used.
        :param params: Optional build parameters.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        :type params: dict
        :type parallel: int

        Endpoint:
            POST: ``project/:vcs-type/:username/:project/tree/:branch``
        """
        data = {
            'revision': revision,
            'tag': tag,
            'parallel': parallel,
        }

        if params:
            data.update(params)

        endpoint = 'project/{0}/{1}/{2}/tree/{3}'.format(
            vcs_type,
            username,
            project,
            branch
        )

        resp = self._request('POST', endpoint, data=data)
        return resp

    def add_ssh_key(
            self,
            username,
            project,
            ssh_key,
            vcs_type='github',
            hostname=None):
        """Create an ssh key

        Used to access external systems that require SSH key-based authentication.

        .. note::
            The ssh_key must be unencrypted.

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param branch: Defaults to master.
        :param ssh_key: Private RSA key.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.
        :param hostname: Optional hostname. If set, the key will only work \
            for this hostname.

        Endpoint:
            POST: ``/project/:vcs-type/:username/:project/ssh-key``
        """
        endpoint = 'project/{0}/{1}/{2}/ssh-key'.format(
            vcs_type,
            username,
            project
        )

        params = {
            "hostname": hostname,
            "private_key": ssh_key
        }

        resp = self._request('POST', endpoint, data=params)
        return resp

    def list_checkout_keys(self, username, project, vcs_type='github'):
        """List checkout keys for a project

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        Endpoint:
            GET: ``project/:vcs-type/:username/:project/checkout-key``
        """
        endpoint = 'project/{0}/{1}/{2}/checkout-key'.format(
            vcs_type,
            username,
            project
        )

        resp = self._request('GET', endpoint)
        return resp

    def create_checkout_key(self, username, project, key_type, vcs_type='github'):
        """Create a new checkout keys for a project

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param key_type: The type of key to create. Valid values are \
            'deploy-key' or 'github-user-key'
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        :raises InvalidKeyError: When key_type is not a valid key type.

        Endpoint:
            POST: ``/project/:vcs-type/:username/:project/checkout-key``
        """
        valid_types = ['deploy-key', 'github-user-key']

        if key_type not in valid_types:
            raise BadKeyError(key_type)

        params = {
            "type": key_type
        }

        endpoint = 'project/{0}/{1}/{2}/checkout-key'.format(
            vcs_type,
            username,
            project
        )

        resp = self._request('POST', endpoint, data=params)
        return resp

    def get_checkout_key(self, username, project, fingerprint, vcs_type='github'):
        """Get a checkout key.

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param fingerprint: The fingerprint of the checkout key.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        Endpoint:
            GET: ``/project/:vcs-type/:username/:project/checkout-key/:fingerprint``
        """
        endpoint = 'project/{0}/{1}/{2}/checkout-key/{3}'.format(
            vcs_type,
            username,
            project,
            fingerprint
        )

        resp = self._request('GET', endpoint)

        return resp

    def delete_checkout_key(self, username, project, fingerprint, vcs_type='github'):
        """Delete a checkout key.

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param fingerprint: The fingerprint of the checkout key.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        Endpoint:
            DELETE: ``/project/:vcs-type/:username/:project/checkout-key/:fingerprint``
        """
        endpoint = 'project/{0}/{1}/{2}/checkout-key/{3}'.format(
            vcs_type,
            username,
            project,
            fingerprint
        )

        resp = self._request('DELETE', endpoint)
        return resp

    def get_test_metadata(self, username, project, build_num, vcs_type='github'):
        """Provides test metadata for a build

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param build_num: Build number.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        Endpoint:
            GET: ``/project/:vcs-type/:username/:project/:build_num/tests``
        """
        endpoint = 'project/{0}/{1}/{2}/{3}/tests'.format(
            vcs_type,
            username,
            project,
            build_num
        )

        resp = self._request('GET', endpoint)
        return resp

    def list_envvars(self, username, project, vcs_type='github'):
        """Provides list of environment variables for a project

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        Endpoint:
            GET: ``/project/:vcs-type/:username/:project/envvar``
        """
        endpoint = 'project/{0}/{1}/{2}/envvar'.format(
            vcs_type,
            username,
            project
        )

        resp = self._request('GET', endpoint)
        return resp

    def add_envvar(self, username, project, name, value, vcs_type='github'):
        """Adds an environment variable to a project

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param name: Name of the environment variable.
        :param value: Value of the environment variable.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        Endpoint:
            POST: ``/project/:vcs-type/:username/:project/envvar``
        """
        params = {
            "name": name,
            "value": value
        }

        endpoint = 'project/{0}/{1}/{2}/envvar'.format(
            vcs_type,
            username,
            project
        )

        resp = self._request('POST', endpoint, data=params)
        return resp

    def get_envvar(self, username, project, name, vcs_type='github'):
        """Gets the hidden value of an environment variable

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param name: Name of the environment variable.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        Endpoint:
            GET ``/project/:vcs-type/:username/:project/envvar/:name``
        """
        endpoint = 'project/{0}/{1}/{2}/envvar/{3}'.format(
            vcs_type,
            username,
            project,
            name
        )

        resp = self._request('GET', endpoint)

        return resp

    def delete_envvar(self, username, project, name, vcs_type='github'):
        """Delete an environment variable

        :param username: Org or user name.
        :param project: Case sensitive repo name.
        :param name: Name of the environment variable.
        :param vcs_type: Defaults to github. On circleci.com you can \
            also pass in ``bitbucket``.

        Endpoint:
            DELETE ``/project/:vcs-type/:username/:project/envvar/:name``
        """
        endpoint = 'project/{0}/{1}/{2}/envvar/{3}'.format(
            vcs_type,
            username,
            project,
            name
        )

        resp = self._request('DELETE', endpoint)

        return resp

    def _request(self, verb, endpoint, data=None):
        """Request a url.

        :param endpoint: The api endpoint we want to call.
        :param verb: POST, GET, or DELETE.
        :param params: Optional build parameters.

        :type params: dict

        :raises requests.exceptions.HTTPError: When response code is not successful.

        :returns: A JSON object with the response from the API.
        """

        headers = {
            'Accept': 'application/json',
        }
        auth = HTTPBasicAuth(self.token, '')
        resp = None

        request_url = "{0}/{1}".format(self.url, endpoint)

        if verb == 'GET':
            resp = requests.get(request_url, auth=auth, headers=headers)
        elif verb == 'POST':
            resp = requests.post(request_url, auth=auth, headers=headers, json=data)
        elif verb == 'DELETE':
            resp = requests.delete(request_url, auth=auth, headers=headers)
        else:
            raise BadVerbError(verb)

        resp.raise_for_status()

        return resp.json()

    def _download(self, url, destdir=None, filename=None):
        """File download helper.

        :param url: The URL to the artifact.
        :param destdir: The optional destination directory. \
            Defaults to None (curent working directory).
        :param filename: Optional file name. Defaults to the name of the artifact file.
        """
        if not filename:
            filename = url.split('/')[-1]

        if not destdir:
            destdir = os.getcwd()

        endpoint = "{0}?circle-token={1}".format(url, self.token)

        resp = requests.get(endpoint, stream=True)

        path = "{0}/{1}".format(destdir, filename)

        with open(path, 'wb') as f:
            for chunk in resp.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)

        return path


import yaml
import requests
import jwt
import json
import uuid
import os,sys, traceback
import datetime
import pickledb


def openFile(fileName):
    data = ""
    with open (fileName, "r") as f:
        data = f.read()
    return data

def createToken(thingId, payload, privateKey):
    return jwt.encode(payload, privateKey, algorithm='RS256', headers={'kid': thingId}).decode('utf-8')
    
class FlashlexSDK(object):

    def __init__(self, config):
        if(type(config) == 'str'):
            self._config = self.loadConfig(config)
        else:
            self._config = config
       
    def getConfig(self):
        return self._config
    
    def setConfig(self, config):
        self._config = config
    
    def loadConfig(self, configFile):
        cfg = None
        with open(configFile, 'r') as ymlfile:
            cfg = yaml.load(ymlfile, Loader=yaml.FullLoader)
        return cfg

    def getSubscribedMessages(self, count=-1, sortMessages=True, reverse=False):
        subscriptionDataPath = "{0}/{1}".format(
            self._config["flashlex"]["app"]["db"]["dataPath"], 
            self._config["flashlex"]["app"]["db"]["subscriptionData"])

        subscriptionDb = pickledb.load(subscriptionDataPath, False, sig=False)
        listKeys = subscriptionDb.getall()
        messages_all = []
        for key in listKeys:
            messages_all.append(subscriptionDb.get(key))

        print(messages_all)
        if(sortMessages):
            sorted(messages_all, key=lambda message: message["timestamp"], reverse=reverse)
        if(count>0):
            return messages_all[:count]
        else:     
            return messages_all
  
    def removeMessageFromStore(self, message):
        subscriptionDataPath = "{0}/{1}".format(
            self._config["flashlex"]["app"]["db"]["dataPath"], 
            self._config["flashlex"]["app"]["db"]["subscriptionData"])

        if 'pk' in message:
            subscriptionDb = pickledb.load(subscriptionDataPath, False, sig=False)
            subscriptionDb.rem(message['pk'])
            subscriptionDb.dump()
        else:
            raise KeyError('pk key is required')

    def setMessageToStore(self, message):
        subscriptionDataPath = "{0}/{1}".format(
            self._config["flashlex"]["app"]["db"]["dataPath"], 
            self._config["flashlex"]["app"]["db"]["subscriptionData"])

        #should add pk if missing
        if 'pk' not in message:
            message["pk"] = str(uuid.uuid4()).replace("-","")[:12]

        subscriptionDb = pickledb.load(subscriptionDataPath, False, sig=False)
        subscriptionDb.set(message['pk'], message)
        subscriptionDb.dump()


    
    def collectMessage(self, message):
        """
        uses the certificate to generate an asymetric token
        that can be verified by the public key at flashlex,
        messages cannot be collected unless the key is 
        verified.
        """
        thingId=self._config["flashlex"]["thing"]["id"]
        print("collecting message from thing: {0}".format(thingId))
        print(self._config["flashlex"]["thing"]["keys"]["path"])
        privateKey = openFile("{0}/{1}".format(
            self._config["flashlex"]["thing"]["keys"]["path"], 
            self._config["flashlex"]["thing"]["keys"]["privateKey"]))

        
        payload = {
            'exp': datetime.datetime.utcnow() + datetime.timedelta(seconds=360),
            'nbf': datetime.datetime.utcnow() - datetime.timedelta(seconds=360),
            'iss': 'urn:thing:{0}'.format(thingId),
            'aud': 'urn:flashlex:{0}'.format(thingId)
        }

        jwt = createToken(thingId, payload, privateKey)
        
        url = "{0}/{1}/v1/things/{2}/collect".format(
                self._config["flashlex"]["app"]['api']["endpoint"],
                self._config["flashlex"]["app"]['api']["env"],
                self._config["flashlex"]["thing"]["id"])
        # payload = {'key1': 'value1', 'key2': 'value2'}        
        headers={"Authorization":"Bearer {0}".format(jwt), 'Content-Type': 'application/json'}
        print(url, message, headers)

        try:
            r = requests.post(url, 
                data=json.dumps(message), 
                headers=headers)
            print(r.text)    
            return r.status_code
        except:
            print("there was a problem.")
            print("-"*60)
            traceback.print_exc(file=sys.stdout)
            print("-"*60)

            return 500
        


